import requests


def _make_request(url):

    print(url)
    resp = requests.get(url)
    if resp.status_code == 200:
        return resp.text
    elif resp.status_code == 404:
        print(f"Dataset not available (404 Error) for url: \n {url}")
        return None
    else:
        raise ValueError(f"Error code {resp.status_code} for url: \n {url}")



import pandas as pd
import requests
from io import StringIO
from .request import _make_request

BASE_URL = "http://www.ndbc.noaa.gov/data/realtime2"


def ndbc(buoy, dataset="txt"):
    """Get realtime data from the NDBC. They have six different data sources
    to pull from:

    Parameters
    ----------
    buoy : int
        Integer of the buoy identifier. Look up buoy numbers here:
        https://www.ndbc.noaa.gov/
    source: str
        What type of data to query. Possible values are:
        - data_spec: Raw Spectral Wave Data
        - ocean: Oceanographic Data
        - spec: Spectral Wave Summary Data
        - supl: Supplemental Measurements Data
        - swdir: Spectral Wave Data (alpha1)
        - swdir2: Spectral Wave Data (alpha2)
        - swr1: Spectral Wave Data (r1)
        - swr2: Spectral Wave Data (r2)
        - txt: Standard Meteorological Data

    Returns
    -------
    df : pd.DataFrame
        DataFrame containing the requested data.

    Notes
    -----

    The realtime data for all of their buoys can be found at:
    http://www.ndbc.noaa.gov/data/realtime2/

    Info about all of noaa data can be found at:
    http://www.ndbc.noaa.gov/docs/ndbc_web_data_guide.pdf

    What all the values mean:
    http://www.ndbc.noaa.gov/measdes.shtml


    Example:
    from seebuoy import ndbc

    # Get the last 45 days of data
    df = ndbc(41013, 'txt')
    df.head()

    Out[7]:
                        WVHT    SwH SwP WWH WWP SwD WWD STEEPNESS   APD MWD
    2016-02-04 17:42:00 1.6     1.3 7.1 0.9 4.5 S   S   STEEP       5.3 169
    2016-02-04 16:42:00 1.7     1.5 7.7 0.9 5.0 S   S   STEEP       5.4 174
    2016-02-04 15:41:00 2.0     0.0 NaN 2.0 7.1 NaN S   STEEP       5.3 174
    2016-02-04 14:41:00 2.0     1.2 7.7 1.5 5.9 SSE SSE STEEP       5.5 167
    2016-02-04 13:41:00 2.0     1.7 7.1 0.9 4.8 S   SSE STEEP       5.7 175
    """

    opts = {
        "data_spec": _data_spec,
        "ocean": _ocean,
        "spec": _spec,
        "supl": _supl,
        "swdir": _swdir,
        "swdir2": _swdir2,
        "swr1": _swr1,
        "swr2": _swr2,
        "txt": _txt,
    }

    if dataset not in opts:
        raise ValueError("Dataset must be one of {}".format(", ".join(opts)))

    url = f"{BASE_URL}/{buoy}.{dataset}"
    txt = _make_request(url)

    if txt is None:
        return

    return opts[dataset](txt)


def _data_spec(txt):

    df = pd.read_csv(
        StringIO(txt),
        delim_whitespace=True,
        skiprows=1,
        header=None,
        parse_dates=[[0, 1, 2, 3, 4]],
        index_col=0,
    )

    # convert the dates to datetimes
    df.index = pd.to_datetime(df.index, format="%Y %m %d %H %M")
    df.index.name = "date"

    specs = df.iloc[:, 1::2]
    freqs = df.iloc[0, 2::2]

    specs.columns = freqs

    # remove the parenthesis from the column index
    specs.columns = [c.replace("(", "").replace(")", "") for c in specs.columns]

    return specs


def _ocean(txt):

    df = pd.read_csv(
        StringIO(txt),
        delim_whitespace=True,
        na_values="MM",
        parse_dates=[[0, 1, 2, 3, 4]],
        index_col=0,
    )

    # units are in the second row drop them
    df.drop(df.index[0], inplace=True)

    # convert the dates to datetimes
    df.index = pd.to_datetime(df.index, format="%Y %m %d %H %M")
    df.index.name = "date"

    # convert to floats
    cols = ["DEPTH", "OTMP", "COND", "SAL"]
    df[cols] = df[cols].astype(float)

    return df


def _spec(txt):

    df = pd.read_csv(
        StringIO(txt),
        delim_whitespace=True,
        na_values="MM",
        parse_dates=[[0, 1, 2, 3, 4]],
        index_col=0,
    )

    try:
        # units are in the second row drop them
        # df.columns = df.columns + '('+ df.iloc[0] + ')'
        df.drop(df.index[0], inplace=True)

        # convert the dates to datetimes
        df.index = pd.to_datetime(df.index, format="%Y %m %d %H %M")

        # convert to floats
        cols = ["WVHT", "SwH", "SwP", "WWH", "WWP", "APD", "MWD"]
        df[cols] = df[cols].astype(float)
    except IndexError:

        # convert the dates to datetimes
        df.index = pd.to_datetime(df.index, format="%Y %m %d %H %M")

        # convert to floats
        cols = ["H0", "SwH", "SwP", "WWH", "WWP", "AVP", "MWD"]
        df[cols] = df[cols].astype(float)
    df.index.name = "date"
    return df


def _supl(txt):

    df = pd.read_csv(
        StringIO(txt),
        delim_whitespace=True,
        na_values="MM",
        parse_dates=[[0, 1, 2, 3, 4]],
        index_col=0,
    )

    # units are in the second row drop them
    df.drop(df.index[0], inplace=True)

    # convert the dates to datetimes
    df.index = pd.to_datetime(df.index, format="%Y %m %d %H %M")

    # convert to floats
    cols = ["PRES", "PTIME", "WSPD", "WDIR", "WTIME"]
    for col in cols:
        df[col] = pd.to_numeric(df[col], errors="coerce")

    return df


def _swdir(txt):

    df = pd.read_csv(
        StringIO(txt),
        delim_whitespace=True,
        skiprows=1,
        na_values=999,
        header=None,
        parse_dates=[[0, 1, 2, 3, 4]],
        index_col=0,
    )

    # convert the dates to datetimes
    df.index = pd.to_datetime(df.index, format="%Y %m %d %H %M")
    df.index.name = "date"

    specs = df.iloc[:, 0::2]
    freqs = df.iloc[0, 1::2]

    specs.columns = freqs

    # remove the parenthesis from the column index
    specs.columns = [cname.replace("(", "").replace(")", "") for cname in specs.columns]

    return specs


def _swdir2(txt):

    df = pd.read_csv(
        StringIO(txt),
        delim_whitespace=True,
        skiprows=1,
        header=None,
        parse_dates=[[0, 1, 2, 3, 4]],
        index_col=0,
    )

    # convert the dates to datetimes
    df.index = pd.to_datetime(df.index, format="%Y %m %d %H %M")
    df.index.name = "date"

    specs = df.iloc[:, 0::2]
    freqs = df.iloc[0, 1::2]

    specs.columns = freqs

    # remove the parenthesis from the column index
    specs.columns = [cname.replace("(", "").replace(")", "") for cname in specs.columns]

    return specs


def _swr1(txt):

    df = pd.read_csv(
        StringIO(txt),
        delim_whitespace=True,
        skiprows=1,
        header=None,
        parse_dates=[[0, 1, 2, 3, 4]],
        index_col=0,
        na_values=999,
    )

    # convert the dates to datetimes
    df.index = pd.to_datetime(df.index, format="%Y %m %d %H %M")
    df.index.name = "date"

    specs = df.iloc[:, 0::2]
    freqs = df.iloc[0, 1::2]

    specs.columns = freqs

    # remove the parenthesis from the column index
    specs.columns = [cname.replace("(", "").replace(")", "") for cname in specs.columns]

    return specs


def _swr2(txt):

    df = pd.read_csv(
        StringIO(txt),
        delim_whitespace=True,
        skiprows=1,
        header=None,
        parse_dates=[[0, 1, 2, 3, 4]],
        index_col=0,
        na_values=999,
    )

    # convert the dates to datetimes
    df.index = pd.to_datetime(df.index, format="%Y %m %d %H %M")
    df.index.name = "date"
    specs = df.iloc[:, 0::2]
    freqs = df.iloc[0, 1::2]

    specs.columns = freqs

    # remove the parenthesis from the column index
    specs.columns = [cname.replace("(", "").replace(")", "") for cname in specs.columns]

    return specs


def _txt(txt):

    df = pd.read_csv(
        StringIO(txt),
        delim_whitespace=True,
        na_values="MM",
        parse_dates=[[0, 1, 2, 3, 4]],
        index_col=0,
    )

    # first column is units, so drop it
    df.drop(df.index[0], inplace=True)

    # convert the dates to datetimes
    df.index = pd.to_datetime(df.index, format="%Y %m %d %H %M")
    df.index.name = "date"

    for c in df.columns:
        df[c] = pd.to_numeric(df[c])

    return df



import requests
from io import StringIO
import pandas as pd
from bs4 import BeautifulSoup
from .request import _make_request

BASE_URL = "http://www.ndbc.noaa.gov/view_text_file.php?filename="


def _get_all_urls(buoy):
    """Scrape all the urls for the given buoy. Parses urls at:

    https://www.ndbc.noaa.gov/station_history.php?station=41037

    and finds all the urls that refer to .txt files.
    """

    url = f"https://www.ndbc.noaa.gov/station_history.php?station={buoy}"

    txt = _make_request(url)

    soup = BeautifulSoup(txt)

    urls = soup.find("ul").find_all("a", href=True)
    data_urls = []
    for u in urls:
        if ".txt" in u["href"]:
            data_urls.append(u["href"])

    return data_urls


def available_downloads(buoy):
    """Parse out what the urls actually denote. Date and data type.

    Link examples:
    ex1: /download_data.php?filename=41037h2016.txt.gz&dir=data/historical/stdmet/
    ex2: /download_data.php?filename=4103772020.txt.gz&dir=data/stdmet/Jul/
    ex3: /data/ocean/Aug/41037.txt
    """

    data_urls = _get_all_urls(buoy)
    data_dict = []

    for link in data_urls:
        # remove /historical so they are all the same
        dataset = link.replace("/historical", "").split("data/")[-1].split("/")[0]

        if "filename" in link:
            year = link.split("filename=")[-1].split(".txt")[0][-4:]

            if "historical" not in link:
                month = link.split("/")[-2]
            else:
                month = ""

        else:
            # get max year from previously passed dates.
            year = max([x["year"] for x in data_dict])
            month = link.split("/")[3]

        data_dict.append(
            {"year": year, "month": month, "dataset": dataset, "url": link}
        )
    df = pd.DataFrame(data_dict)

    # Those without a month are assigned Jan for datetime parsing
    m = df["month"] == ""
    df.loc[m, "month"] = "Jan"
    df["date"] = pd.to_datetime(df[["year", "month"]].sum(axis=1), format="%Y%b")

    return df.sort_values(["dataset", "date"])


def ndbc_historic(buoy, year, dataset="stdmet"):
    """Get historical data.

    Parameters
    ----------
    year: int or list of ints
        Years to pull data. Can either be 2008 or [2008, 2009]
    """

    df_avail = available_downloads(buoy)

    parsers = {
        "stdmet": _stdmet,
    }

    if dataset not in df_avail["dataset"].values:
        raise ValueError(
            f"Dataset {dataset} not available for year {year} on buoy {buoy}"
        )

    if dataset not in parsers:
        raise ValueError("Dataset must be one of {}".format(", ".join(parsers)))

    m1 = df_avail["dataset"] == dataset
    m2 = df_avail["year"] == str(year)
    df_store = []
    for url in df_avail.loc[m1 & m2, "url"].values:

        qs = url.split("?")[-1]
        data_url = f"https://www.ndbc.noaa.gov/view_text_file.php?{qs}"
        txt = _make_request(data_url)

        df = parsers[dataset](txt)
        df_store.append(df)

    return pd.concat(df_store)


def _stdmet(txt):

    df = pd.read_csv(
        StringIO(txt),
        header=0,
        delim_whitespace=True,
        na_values=[99, 999, 9999, 99.0, 999.0, 9999.0],
    )

    print(df.head(3))
    # first row is units, so drop it. data after 2007 has units
    if df.iloc[0, 0] == "#yr":
        df = df.drop(df.index[0])

    # data before 2007 does not always have minute
    if "mm" in df.columns:
        res = df.iloc[:, :5].astype(str).agg("-".join, axis=1)
        df["date"] = pd.to_datetime(res, format="%Y-%m-%d-%H-%M")

    else:
        res = df.iloc[:, :4].agg("-".join, axis=1)
        df["date"] = pd.to_datetime(res, format="%Y-%m-%d-%H")

    df = df.set_index("date")

    # cols = [
    #     "WDIR",
    #     "WSPD",
    #     "GST",
    #     "WVHT",
    #     "DPD",
    #     "APD",
    #     "MWD",
    #     "PRES",
    #     "ATMP",
    #     "WTMP",
    #     "DEWP",
    #     "VIS",
    #     "TIDE",
    # ]
    # df = df[cols].astype(float)
    df.columns = df.columns.str.lower()
    return df.astype(float)



import pandas as pd
from datetime import datetime
import numpy as np
import netCDF4


class WW3:
    """Pull wave watch 3 model data in Pandas dataframes.

    Parameters
    ----------
    date : str
        Date of data to pull. If None, it will return the most recent

    """

    def __init__(self, date=None):

        if date is None:
            date = datetime.today().strftime("%Y%m%d")

        base_url = "https://nomads.ncep.noaa.gov:9090/dods/wave/nww3/nww3"
        self.url = "{}{}/nww3{}_00z".format(base_url, date, date)

    def get_file(self):
        """Get the netcdf file constructed from the URL. By default all
        variables are kept.
        """

        self.dataset = netCDF4.Dataset(self.url)

    def get_var(self, var):
        """Get the specified variable from the netcdf file.

        dirpwsfc - primary wave direction [deg]

        dirswsfc - secondary wave direction [deg]

        htsgwsfc- significant height of combined wind waves and swell [m]

        perpwsfc - primary wave mean period [s]

        perswsfc - secondary wave mean period [s]

        ugrdsfc - u-component of wind [m/s]

        vgrdsfc - v-component of wind [m/s]

        wdirsfc - wind direction (from which blowing) [deg]

        windsfc - wind speed [m/s]

        wvdirsfc - direction of wind waves [deg]

        wvpersfc - mean period of wind waves [s]
        Parameters
        ----------
        var : str
            Variable from the netcdf file to retreive.

        Returns
        -------
        x : np.array
            Returns the requested data assiciated with the variable.
        """

        return self.dataset.variables[var].data

    def get_closest_lat_lon(self, lat, lon):
        """Retrieves the indexes of the closest latitude and longitude.

        Parameters
        ----------
        lat : float
            Latitude to find closest of.
        lon : float
            Longitude to find the closest.
        """

        if lon < 0:
            lon = 360 + lon

        xlats = self.dataset.variables["lat"][:].data
        xlons = self.dataset.variables["lon"][:].data

        lat_idx = np.abs(lat - xlats).argmin()
        lon_idx = np.abs(lon - xlons).argmin()

        return lat_idx, lon_idx

    def get_lat_lon_var(self, lat, lon, var):
        """Get the requested variable at the requested latitude and longitude.
        Latitudes range from -80 to 80 and longitudes range from 0 to 360.
        If a negative longitude is fed, it will be converted to a 0 to 360
        range.

        Parameters
        ----------
        lat : float
            Latitude of variable.
        lon : float
            Longitude of variable.
        var : str
            Varibale to retrieve.
        """

        lat_idx, lon_idx = self.get_closest_lat_lon(lat, lon)

        return self.dataset.variables[var].data[1:, lat_idx, lon_idx]

    def get_standard_df(self, lat, lon):
        """Gets the standard data for a specific latitude and longitude.
        Returns it as a dataframe.

        Parameters
        ----------
        lat : float
            Latitude to retrieve for the data
        lon : float
            Longitude to retrieve for the data

        Returns
        -------
        df : pd.DataFrame
            Dataframe of standard values
        """

        today = datetime.today().strftime("%Y-%m-%d")
        dti = pd.date_range(today, periods=60, freq="3H")

        lat_idx, lon_idx = self.get_closest_lat_lon(lat, lon)

        dvars = [
            "dirpwsfc",
            "dirswsfc",
            "htsgwsfc",
            "perpwsfc",
            "perswsfc",
            "ugrdsfc",
            "vgrdsfc",
            "wdirsfc",
            "windsfc",
            "wvdirsfc",
            "wvpersfc",
        ]

        df = {}
        for var in dvars:
            print(var)
            df[var] = self.dataset.variables[var][1:, lat_idx, lon_idx]

        df = pd.DataFrame(df, index=dti)

        return df
